import re
from urllib.parse import urlparse

import biolib.utils
from biolib.typing_utils import Optional, Tuple


def parse_result_id_or_url(result_id_or_url: str, default_token: Optional[str] = None) -> Tuple[str, Optional[str]]:
    result_id_or_url = result_id_or_url.strip()

    if '/' not in result_id_or_url:
        return (result_id_or_url, default_token)

    if not result_id_or_url.startswith('http://') and not result_id_or_url.startswith('https://'):
        result_id_or_url = 'https://' + result_id_or_url

    parsed_url = urlparse(result_id_or_url)

    if biolib.utils.BIOLIB_BASE_URL:
        expected_base = urlparse(biolib.utils.BIOLIB_BASE_URL)
        if parsed_url.scheme != expected_base.scheme or parsed_url.netloc != expected_base.netloc:
            raise ValueError(f'URL must start with {biolib.utils.BIOLIB_BASE_URL}, got: {result_id_or_url}')

    pattern = r'/results?/(?P<uuid>[a-f0-9-]+)/?(?:\?token=(?P<token>[^&]+))?'
    match = re.search(pattern, result_id_or_url, re.IGNORECASE)

    if not match:
        raise ValueError(f'URL must be in format <base_url>/results/<UUID>/?token=<token>, got: {result_id_or_url}')

    uuid = match.group('uuid')
    token = match.group('token') or default_token

    return (uuid, token)



import os

from biolib import utils
from biolib._internal.file_utils import get_files_and_size_of_directory, get_iterable_zip_stream
from biolib._internal.types.typing import List, Optional, Tuple
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger


def validate_data_path_and_get_files_and_size_of_directory(data_path: str) -> Tuple[List[str], int]:
    assert os.path.isdir(data_path), f'The path "{data_path}" is not a directory.'

    if os.path.realpath(data_path) == '/':
        raise BioLibError('Pushing your root directory is not possible')

    original_working_dir = os.getcwd()
    os.chdir(data_path)
    files_to_zip, data_size_in_bytes = get_files_and_size_of_directory(directory=os.getcwd())
    os.chdir(original_working_dir)

    if data_size_in_bytes > 4_500_000_000_000:
        raise BioLibError('Attempted to push directory with a size larger than the limit of 4.5 TB')

    return files_to_zip, data_size_in_bytes


def push_data_path(
    data_path: str,
    data_size_in_bytes: int,
    files_to_zip: List[str],
    resource_version_uuid: str,
    chunk_size_in_mb: Optional[int] = None,
) -> None:
    original_working_dir = os.getcwd()
    os.chdir(data_path)

    min_chunk_size_bytes = 10_000_000
    chunk_size_in_bytes: int
    if chunk_size_in_mb:
        chunk_size_in_bytes = chunk_size_in_mb * 1_000_000  # Convert megabytes to bytes
        if chunk_size_in_bytes < min_chunk_size_bytes:
            logger.warning('Specified chunk size is too small, using minimum of 10 MB instead.')
            chunk_size_in_bytes = min_chunk_size_bytes
    else:
        # Calculate chunk size based on max chunk count of 10_000, using 9_000 to be on the safe side
        chunk_size_in_bytes = max(min_chunk_size_bytes, int(data_size_in_bytes / 9_000))

    data_size_in_mb = round(data_size_in_bytes / 10**6)
    logger.info(f'Zipping {len(files_to_zip)} files, in total ~{data_size_in_mb}mb of data')

    iterable_zip_stream = get_iterable_zip_stream(files=files_to_zip, chunk_size=chunk_size_in_bytes)
    multipart_uploader = utils.MultiPartUploader(
        use_process_pool=True,
        get_presigned_upload_url_request=dict(
            headers=None,
            requires_biolib_auth=True,
            path=f'/lfs/versions/{resource_version_uuid}/presigned_upload_url/',
        ),
        complete_upload_request=dict(
            headers=None,
            requires_biolib_auth=True,
            path=f'/lfs/versions/{resource_version_uuid}/complete_upload/',
        ),
    )

    multipart_uploader.upload(payload_iterator=iterable_zip_stream, payload_size_in_bytes=data_size_in_bytes)
    os.chdir(original_working_dir)



import time

from biolib.biolib_binary_format.utils import IndexableBuffer
from biolib.biolib_logging import logger
from biolib.typing_utils import Iterable


class StreamSeeker:
    def __init__(
            self,
            upstream_buffer: IndexableBuffer,
            files_data_start: int,
            files_data_end: int,
            max_chunk_size: int,
    ):
        self._upstream_buffer = upstream_buffer
        self._files_data_end = files_data_end
        self._download_chunk_size_in_bytes = 100_000
        self._min_chunk_size = min(100_000, max_chunk_size)
        self._max_chunk_size = max_chunk_size
        self._target_download_time_seconds = 1.0

        self._buffer_start = files_data_start
        self._buffer = bytearray()

    def seek_and_read(self, file_start: int, file_length: int) -> Iterable[bytes]:
        assert file_start >= self._buffer_start
        self._buffer = self._buffer[file_start - self._buffer_start:]  # Returns empty array if "out of bounds"
        self._buffer_start = file_start

        while True:
            file_byte_count_remaining = file_length - (self._buffer_start - file_start)
            if file_byte_count_remaining == 0:
                return

            start_of_fetch = self._buffer_start + len(self._buffer)
            byte_count_left_in_stream = self._files_data_end - start_of_fetch

            if byte_count_left_in_stream != 0:
                # Only fetch if there is still data left upstream
                if self._download_chunk_size_in_bytes > len(self._buffer):
                    # Only fetch if size of buffer is below chunk size
                    fetch_size = min(byte_count_left_in_stream, self._download_chunk_size_in_bytes)

                    start_time = time.monotonic()
                    fetched_data = self._upstream_buffer.get_data(
                        start=start_of_fetch,
                        length=fetch_size,
                    )
                    download_time = time.monotonic() - start_time

                    self._buffer.extend(fetched_data)

                    if download_time > 0:
                        self._adjust_chunk_size(download_time, fetch_size)

            bytes_to_yield = self._buffer[:file_byte_count_remaining]  # Returns empty array if "out of bounds"
            yield bytes_to_yield
            self._buffer = self._buffer[file_byte_count_remaining:]  # Returns empty array if "out of bounds"
            self._buffer_start += len(bytes_to_yield)

    def _adjust_chunk_size(self, download_time: float, _bytes_downloaded: int) -> None:
        new_chunk_size = self._download_chunk_size_in_bytes
        time_ratio = download_time / self._target_download_time_seconds

        if time_ratio > 1.1:
            adjustment_factor = 1.0 / time_ratio
            adjustment_factor = max(adjustment_factor, 0.5)
            new_chunk_size = int(self._download_chunk_size_in_bytes * adjustment_factor)
        elif time_ratio < 0.9:
            adjustment_factor = 1.0 / time_ratio
            new_chunk_size = int(self._download_chunk_size_in_bytes * adjustment_factor)

        new_chunk_size = max(
            self._min_chunk_size,
            min(self._max_chunk_size, new_chunk_size)
        )

        if new_chunk_size != self._download_chunk_size_in_bytes:
            logger.debug(
                f"Adjusting chunk size: {self._download_chunk_size_in_bytes} -> {new_chunk_size} bytes "
                f"(download_time={download_time:.2f}s, time_ratio={time_ratio:.2f})"
            )
            self._download_chunk_size_in_bytes = new_chunk_size



from biolib import utils
from biolib._internal.types import Optional
from biolib.api.client import ApiClient, ApiClientInitDict
from biolib.app import BioLibApp


class Session:
    def __init__(self, _init_dict: ApiClientInitDict, _experiment: Optional[str] = None) -> None:
        self._api = ApiClient(_init_dict=_init_dict)
        self._experiment = _experiment

    @staticmethod
    def get_session(refresh_token: str, base_url: Optional[str] = None, client_type: Optional[str] = None, experiment: Optional[str] = None) -> 'Session':
        return Session(
            _init_dict=ApiClientInitDict(
                refresh_token=refresh_token,
                base_url=base_url or utils.load_base_url_from_env(),
                client_type=client_type,
            ),
            _experiment=experiment,
        )

    def load(self, uri: str, suppress_version_warning: bool = False) -> BioLibApp:
        r"""Load a BioLib application by its URI or website URL.

        Args:
            uri (str): The URI or website URL of the application to load. Can be either:
                - App URI (e.g., 'biolib/myapp:1.0.0')
                - Website URL (e.g., 'https://biolib.com/biolib/myapp/')
            suppress_version_warning (bool): If True, don't print a warning when no version is specified.
                Defaults to False.

        Returns:
            BioLibApp: The loaded application object

        Example::

            >>> # Load by URI
            >>> app = biolib.load('biolib/myapp:1.0.0')
            >>> # Load by website URL
            >>> app = biolib.load('https://biolib.com/biolib/myapp/')
            >>> result = app.cli('--help')
        """
        return BioLibApp(uri=uri, _api_client=self._api, suppress_version_warning=suppress_version_warning, _experiment=self._experiment)



from docker.errors import ImageNotFound
from docker.models.images import Image

from biolib import utils
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_logging import logger_no_user_data
from biolib.typing_utils import cast


def get_biolib_nginx_proxy_image() -> Image:
    docker = BiolibDockerClient().get_docker_client()

    if utils.IS_RUNNING_IN_CLOUD:
        try:
            logger_no_user_data.debug('Getting local Docker image for nginx proxy')
            return cast(Image, docker.images.get('biolib-remote-host-proxy:latest'))
        except ImageNotFound:
            logger_no_user_data.debug(
                'Local Docker image for nginx proxy not available. Falling back to public image...'
            )

    public_image_uri = 'public.ecr.aws/h5y4b3l1/biolib-remote-host-proxy:latest'
    try:
        logger_no_user_data.debug('Getting public Docker image for nginx proxy')
        return cast(Image, docker.images.get(public_image_uri))
    except ImageNotFound:
        logger_no_user_data.debug('Pulling public Docker image for nginx proxy')
        return cast(Image, docker.images.pull(public_image_uri))



from typing import Optional


class BioLibError(Exception):

    def __init__(self, message):
        super().__init__(message)
        self.message = message

class ValidationError(BioLibError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(message=message or 'Invalid input.')

class NotFound(BioLibError):

    def __init__(self, message: Optional[str] = None):
        super().__init__(message=message or 'Not found.')


class DockerContainerNotFoundDuringExecutionException(Exception):
    pass


class RetryLimitException(Exception):
    pass


class StorageDownloadFailed(Exception):
    pass


class CloudJobFinishedError(Exception):
    pass


class JobResultError(BioLibError):
    pass


class JobResultNotFound(JobResultError):
    pass


class JobResultPermissionError(JobResultError):
    pass


class JobResultNonZeroExitCodeError(JobResultError):
    """Raised when an app returns a non-zero exit code and check=True."""
    def __init__(self, exit_code: int, message: Optional[str] = None):
        self.exit_code = exit_code
        super().__init__(
            message or f'App returned non-zero exit code: {exit_code}'
        )



import glob
import os
import re
import shutil
import subprocess
import tempfile

import biolib
from biolib.utils import SeqUtil


def natsorted(lst):
    """Sort the list using the natural sort key."""

    def _natural_sort_key(s):
        """A key function for natural sorting."""
        return [int(text) if text.isdigit() else text.lower() for text in re.split('([0-9]+)', s)]

    return sorted(lst, key=_natural_sort_key)


def fasta_above_threshold(fasta_file, work_threshold, work_per_residue=1, verbose=False):
    """True if total FASYA residue work above max_work"""

    records = SeqUtil.parse_fasta(fasta_file)

    # Calculate work units
    total_work_units = 0
    for i, record in enumerate(records):
        sequence_work_units = len(record.sequence) * work_per_residue
        total_work_units += sequence_work_units

        if total_work_units >= work_threshold:
            if verbose:
                print(f'FASTA above threshold (stopped at {total_work_units}) >= {work_threshold}')
                print(f'From  from {i+1}/{len(records)} sequences in {fasta_file}')
            return True

    if verbose:
        print(f'FASTA below threshold ({total_work_units}) < {work_threshold}')
        print(f'From {len(records)} sequences in {fasta_file}')

    return False


def run_locally(command_list, args):
    """Run script locally (no multi-node processing)"""

    # Prepare command
    new_args = vars(args)

    # Delete multinode-specific input arguments
    for k in list(new_args.keys()):
        if str(k).startswith('multinode'):
            del new_args[k]

    # Convert to list format
    new_args_list = _args_dict_to_args_list(new_args)

    # Prepare command, e.g. ["python3", "predict.py"] + new_args_list
    command = command_list + new_args_list

    if args.verbose >= 1:
        print(f'Running {command}')

    # Run command
    result = subprocess.run(command, capture_output=True, text=True, check=False)
    if result.returncode == 0:
        print(f'{result.stdout}')
    else:
        print(f'Error: {result.stderr}')


def fasta_batch_records(records, work_per_batch_min, work_per_residue=1, verbose=False):
    """Converts FASTA records to batches of records, based on thresholds"""

    def log_batches(batches):
        for i, batch in enumerate(batches):
            batch_dict = {
                'records': len(batch),
                'residues': sum(len(record.sequence) for record in batch),
            }

            n_seqs, n_res = batch_dict['records'], batch_dict['residues']
            print(f'Batch {i+1}: {n_res} residues from {n_seqs} sequences')

    batches = []
    batch = []
    current_longest_seq_len = 0
    for record in records:
        seq_len = len(record.sequence)
        potential_longest_seq_len = max(current_longest_seq_len, seq_len)

        # Calculate work units if we were to add this record
        potential_work_units = potential_longest_seq_len * work_per_residue * (len(batch) + 1)

        if potential_work_units >= work_per_batch_min and len(batch) > 0:
            batches.append(batch)
            batch = []
            current_longest_seq_len = 0
            potential_longest_seq_len = seq_len

        # Add to batch
        batch.append(record)
        current_longest_seq_len = potential_longest_seq_len

    # Append last batch if present
    if batch:
        batches.append(batch)

    if verbose:
        log_batches(batches)

    return batches


def fasta_send_batches_biolib(
    app_url, batches, args, args_fasta='fasta', machine='cpu.large', stream_all_jobs=True, verbose=1
):
    """
    Send jobs through pybiolib interface
    """

    if args.verbose >= 1:
        print(f'Sending {len(batches)} batches to Biolib')

    # Login to biolib, prepare app
    # current_app = biolib.load(Runtime.get_app_uri())
    current_app = biolib.load(app_url)  # Nb: uses "_" not "-"

    # Compute results
    job_list = []
    for i, batch_records in enumerate(batches):  # MH
        # Write FASTA, send to server
        with tempfile.TemporaryDirectory() as tempdir:
            # New arguments
            new_args = vars(args)

            # Write batched FASTA to send
            fasta_path = f'{tempdir}/input.fasta'
            SeqUtil.write_records_to_fasta(fasta_path, batch_records)
            new_args[args_fasta] = fasta_path
            new_args['multinode_only_local'] = True

            # Convert to list
            new_args_list = _args_dict_to_args_list(new_args)

            # Send job
            job = current_app.cli(args=new_args_list, blocking=False, machine=machine)
            job_list.append(job)

            # Job stats
            if args.verbose:
                batch_dict = _get_batch_stats(batch_records)
                n_seqs, n_res = batch_dict['records'], batch_dict['residues']
                print(f'Sending job {i+1}: {n_res} residues from {n_seqs} sequences -> arg_list = {new_args_list}')

    # Stream job output at a time
    print('Streaming job outputs ...')
    for i, job in enumerate(job_list):
        # Try to print if verbose. Always on first job, otherwise only if stream_all_jobs set
        if (i == 0 and verbose) or (stream_all_jobs and verbose):
            job.stream_logs()

        # Check if job succeeded
        assert job.get_exit_code() == 0, f'Job failed with exit code {job.get_exit_code()}'

        # Write to disk
        output_dir = f'job_output/job_{i+1}'
        job.save_files(output_dir=output_dir)

        if verbose:
            print(f'Saving to {output_dir}')


def merge_folder(folder_name, job_out_dir='job_output', out_dir='output', verbose=1):
    """Helper function for merging folders"""

    os.makedirs(out_dir, exist_ok=True)

    job_dirs = glob.glob(f'{job_out_dir}/job_*')
    job_dirs = natsorted(job_dirs)

    # Move first file, prepare to merge
    first_folder = f'{job_dirs[0]}/{folder_name}'
    merged_folder = f'{out_dir}/{folder_name}'
    shutil.move(first_folder, merged_folder)

    if verbose:
        print(f'Merging {folder_name} from {len(job_dirs)} directories to {merged_folder}')

    # If more than one folder, merge to first
    if len(job_dirs) >= 2:
        # Find each job output file
        for job_dir in job_dirs[1:]:
            # Move over extra files
            extra_folder = f'{job_dir}/{folder_name}'
            extra_files = os.listdir(extra_folder)
            for file_name in extra_files:
                file_path = f'{extra_folder}/{file_name}'
                shutil.move(file_path, merged_folder)


def merge_file(
    file_name,
    header_lines_int=1,
    job_out_dir='job_output',
    out_dir='output',
    verbose=1,
):
    """Helper function for merging files with headers"""

    os.makedirs(out_dir, exist_ok=True)

    job_dirs = glob.glob(f'{job_out_dir}/job_*')
    job_dirs = natsorted(job_dirs)

    # Move first file, prepare to merge
    first_file = f'{job_dirs[0]}/{file_name}'
    merged_file = f'{out_dir}/{file_name}'
    shutil.move(first_file, merged_file)

    if verbose:
        print(f'Merging {file_name} from {len(job_dirs)} directories to {merged_file}')

    # If more than one file, append to first
    if len(job_dirs) >= 2:
        # Open first file
        with open(merged_file, 'a') as merged_file_handle:
            # Find each job output file
            for job_dir in job_dirs[1:]:
                # Open extra file
                extra_file = f'{job_dir}/{file_name}'
                with open(extra_file) as extra_file_handle:
                    # Skip first n header lines
                    for _ in range(header_lines_int):
                        next(extra_file_handle)

                    # Append content to first file
                    contents = extra_file_handle.read()
                    merged_file_handle.write(contents)


def _get_batch_stats(batch):
    stats_dict = {
        'records': len(batch),
        'residues': sum(len(R.sequence) for R in batch),
    }

    return stats_dict


def _args_dict_to_args_list(new_args):
    """Converts args dict to list of arguments for Biolib"""

    nested_list = [[f'--{key}', f'{value}'] for key, value in new_args.items()]

    arg_list = []
    for lst in nested_list:
        for item in lst:
            arg_list.append(item)

    return arg_list



import argparse

from biolib.sdk import Runtime

parser = argparse.ArgumentParser(description='Process some biological sequences.')
parser.add_argument('--input', type=str, required=True, help='Input protein sequences')
args = parser.parse_args()

# update the BioLib result name based on the provided file
Runtime.set_result_name_from_file(args.input)

print(f'Processing input file {args.input}...')



import sys

# import and expose everything from the typing module
from typing import *  # noqa:F403 pylint: disable=wildcard-import, unused-wildcard-import

if sys.version_info < (3, 8):  # noqa: UP036
    from typing_extensions import Literal, TypedDict  # pylint: disable=unused-import

if sys.version_info < (3, 11):  # noqa: UP036
    from typing_extensions import NotRequired  # pylint: disable=unused-import



import base64
import os
import random
import shutil
import socket
import sys
import threading
import time
from queue import Queue
from typing import Optional

from biolib import api, utils
from biolib.biolib_binary_format import ModuleOutputV2, SystemException, SystemStatusUpdate
from biolib.biolib_binary_format.utils import LocalFileIndexableBuffer
from biolib.biolib_logging import logger, logger_no_user_data
from biolib.compute_node.cloud_utils import CloudUtils
from biolib.compute_node.job_worker import JobWorkerProcess
from biolib.compute_node.job_worker.job_storage import JobStorage
from biolib.compute_node.socker_listener_thread import SocketListenerThread
from biolib.compute_node.socket_sender_thread import SocketSenderThread
from biolib.compute_node.utils import SystemExceptionCodes, WorkerThreadException, get_package_type
from biolib.compute_node.webserver import webserver_utils

SOCKET_HOST = '127.0.0.1'


class WorkerThread(threading.Thread):
    def __init__(self, compute_state):
        try:
            super().__init__()
            self.compute_state = compute_state
            self._socket_port = random.choice(range(6000, 65000))
            self._socket = None
            self._connection = None
            self._job_worker_process = None
            self._connection_thread = None
            self._listener_thread = None
            self._sender_thread = None
            self._start_and_connect_to_compute_process()

            logger.debug(f'WorkerThread connected to port {self._socket_port}')

        except Exception as exception:
            logger_no_user_data.error(exception)
            raise WorkerThreadException(
                exception,
                SystemExceptionCodes.FAILED_TO_INITIALIZE_WORKER_THREAD.value,
                worker_thread=self,
            ) from exception

    @property
    def _job_uuid(self):
        return self.compute_state['job_id']

    @property
    def _job_temporary_dir(self):
        return self.compute_state['job_temporary_dir']

    def _upload_module_output_and_get_exit_code(self) -> Optional[int]:
        exit_code = None
        try:
            module_output_path = os.path.join(
                self._job_temporary_dir,
                JobStorage.module_output_file_name,
            )
            if os.path.exists(module_output_path):
                module_output = ModuleOutputV2(buffer=LocalFileIndexableBuffer(filename=module_output_path))
                exit_code = module_output.get_exit_code()
                logger_no_user_data.debug(f'Got exit code: {exit_code}')
                if utils.IS_RUNNING_IN_CLOUD:
                    JobStorage.upload_module_output(
                        job_temporary_dir=self._job_temporary_dir,
                        job_uuid=self._job_uuid,
                    )
        except Exception as error:
            logger_no_user_data.error(f'Could not upload module output or get exit code: {error}')
        return exit_code

    def run(self):
        try:
            while True:
                package = self.compute_state['received_messages_queue'].get()
                if package == b'CANCEL_JOB':
                    logger_no_user_data.info(f'Job "{self._job_uuid}" got cancel signal')
                    self.compute_state['status']['error_code'] = SystemExceptionCodes.CANCELLED_BY_USER.value
                    self.terminate()

                package_type = get_package_type(package)

                if package_type == 'StdoutAndStderr':
                    self.compute_state['status']['stdout_and_stderr_packages_b64'].append(
                        base64.b64encode(package).decode()
                    )

                elif package_type == 'SystemStatusUpdate':
                    progress, log_message = SystemStatusUpdate(package).deserialize()
                    self._set_status_update(progress, log_message)

                    # If 'Computation Finished'
                    if progress == 94:
                        self.compute_state['exit_code'] = self._upload_module_output_and_get_exit_code()
                        self._set_status_update(progress=95, log_message='Result Ready')
                        self.compute_state['is_completed'] = True
                        self.terminate()

                elif package_type == 'SystemException':
                    error_code = SystemException(package).deserialize()
                    self.compute_state['status']['error_code'] = error_code
                    logger.debug('Hit error. Terminating Worker Thread and Compute Process')
                    self.compute_state['progress'] = 95
                    self.terminate()

                elif package_type == 'AesEncryptedPackage':
                    if self.compute_state['progress'] == 94:  # Check if encrypted package is ModuleOutput
                        self.compute_state['result'] = package
                        self.terminate()
                    else:  # Else it is StdoutAndStderr
                        self.compute_state['status']['stdout_and_stderr_packages_b64'].append(
                            base64.b64encode(package).decode()
                        )

                else:
                    raise Exception(f'Package type from child was not recognized: {package}')

                self.compute_state['received_messages_queue'].task_done()

        except Exception as exception:
            raise WorkerThreadException(
                exception,
                SystemExceptionCodes.FAILED_TO_HANDLE_PACKAGE_IN_WORKER_THREAD.value,
                worker_thread=self,
            ) from exception

    def _set_status_update(self, progress: int, log_message: str) -> None:
        status_update = dict(progress=progress, log_message=log_message)
        logger_no_user_data.debug(f'Job "{self._job_uuid}" got system log: {status_update}')

        self.compute_state['progress'] = progress
        self.compute_state['status']['status_updates'].append(status_update)

    def _start_and_connect_to_compute_process(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        logger_no_user_data.debug(f'Trying to bind to socket on {SOCKET_HOST}:{self._socket_port}')
        self._socket.bind((SOCKET_HOST, self._socket_port))

        logger_no_user_data.debug(f'Starting to listen to socket on port {self._socket_port}')
        self._socket.listen()
        logger_no_user_data.debug(f'Listening to port {self._socket_port}')

        received_messages_queue = Queue()
        messages_to_send_queue = Queue()

        # Starting a thread for accepting connections before starting the process that should to connect to the socket
        logger_no_user_data.debug('Starting connection thread')
        self._connection_thread = threading.Thread(
            target=self._accept_new_socket_connection,
            args=[received_messages_queue, messages_to_send_queue],
        )
        self._connection_thread.start()
        logger_no_user_data.debug('Started connection thread')
        logger_no_user_data.debug('Starting compute process')

        self._job_worker_process = JobWorkerProcess(socket_port=self._socket_port, log_level=logger.level)
        self._job_worker_process.start()

        self.compute_state['received_messages_queue'] = received_messages_queue
        self.compute_state['messages_to_send_queue'] = messages_to_send_queue
        self.compute_state['worker_thread'] = self

    def _accept_new_socket_connection(self, received_messages_queue, messages_to_send_queue):
        self._connection, _ = self._socket.accept()
        self._listener_thread = SocketListenerThread(self._connection, received_messages_queue)
        self._listener_thread.start()

        self._sender_thread = SocketSenderThread(self._connection, messages_to_send_queue)
        self._sender_thread.start()

    def terminate(self) -> None:
        cloud_job_uuid = self.compute_state['cloud_job_id']
        system_exception_code = self.compute_state['status'].get('error_code')

        if utils.IS_RUNNING_IN_CLOUD and system_exception_code not in [
            SystemExceptionCodes.CANCELLED_BY_USER.value,
            SystemExceptionCodes.EXCEEDED_MAX_JOB_RUNTIME.value,
        ]:
            CloudUtils.finish_cloud_job(
                cloud_job_id=cloud_job_uuid,
                system_exception_code=system_exception_code,
                exit_code=self.compute_state.get('exit_code', None),
            )

        deregistered_due_to_error = False
        if self._job_worker_process:
            logger_no_user_data.debug(
                f'Job "{self._job_uuid}" terminating JobWorkerProcess with PID {self._job_worker_process.pid}'
            )
            self._job_worker_process.terminate()

            clean_up_timeout_in_seconds = 600
            for _ in range(clean_up_timeout_in_seconds):
                if self._job_worker_process.exitcode is not None:
                    logger_no_user_data.debug(
                        f'Job "{self._job_uuid}" worker process exitcode {self._job_worker_process.exitcode}'
                    )
                    break
                else:
                    logger_no_user_data.debug(f'Job "{self._job_uuid}" waiting for worker process to exit...')
                    time.sleep(1)

            if self._job_worker_process.exitcode is None:
                # TODO: Figure out if more error handling is necessary here
                logger_no_user_data.error(
                    f'Job {self._job_uuid} worker process did not exit within {clean_up_timeout_in_seconds} seconds'
                )
                if utils.IS_RUNNING_IN_CLOUD:
                    logger_no_user_data.error('Deregistering compute node...')
                    CloudUtils.deregister(error='job_cleanup_timed_out')
                    deregistered_due_to_error = True

            elif system_exception_code in [
                SystemExceptionCodes.CANCELLED_BY_USER.value,
                SystemExceptionCodes.EXCEEDED_MAX_JOB_RUNTIME.value,
            ]:
                self.compute_state['exit_code'] = self._upload_module_output_and_get_exit_code()
                CloudUtils.finish_cloud_job(
                    cloud_job_id=cloud_job_uuid,
                    system_exception_code=system_exception_code,
                    exit_code=self.compute_state.get('exit_code', None),
                )

        # Delete result as error occurred
        if system_exception_code and os.path.exists(self._job_temporary_dir):
            shutil.rmtree(self._job_temporary_dir)

        if self._socket:
            self._socket.close()

        if self._connection:
            self._connection.close()

        if self.compute_state['progress'] == 95:
            seconds_to_sleep = 5
            logger_no_user_data.debug(
                f'Job "{self._job_uuid}" worker thread sleeping for {seconds_to_sleep} seconds before cleaning up'
            )
            # sleep to let the user start downloading the result
            time.sleep(seconds_to_sleep)

        compute_state_dict = webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT
        if self._job_uuid in compute_state_dict:
            # Delete result as user has not started download
            if compute_state_dict[self._job_uuid]['progress'] == 95 and os.path.exists(self._job_temporary_dir):
                shutil.rmtree(self._job_temporary_dir)

            webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT.pop(self._job_uuid)
            logger_no_user_data.debug(f'Job "{self._job_uuid}" was cleaned up')
        else:
            logger_no_user_data.debug(
                f'Job "{self._job_uuid}" could not be found, maybe it has already been cleaned up'
            )

        if utils.IS_RUNNING_IN_CLOUD:
            config = CloudUtils.get_webserver_config()
            logger_no_user_data.debug(f'Job "{self._job_uuid}" reporting CloudJob "{cloud_job_uuid}" as cleaned up...')
            api.client.post(
                path=f'/internal/compute-nodes/cloud-jobs/{cloud_job_uuid}/cleaned-up/',
                headers={'Compute-Node-Auth-Token': config['compute_node_info']['auth_token']},
            )

            if deregistered_due_to_error:
                CloudUtils.shutdown()  # shutdown now
            else:
                webserver_utils.update_auto_shutdown_time()

        logger_no_user_data.debug(f'Job "{self._job_uuid}" worker thread exiting...')
        sys.exit()
